import 'package:flutter/material.dart';
import 'package:paginated_items_builder/paginated_items_builder.dart';

dynamic _getByType<T>([String? mockItemKey]) => null;

String _noItemsTextGetter(String name) {
  final beforeCapitalLetter = RegExp(r"(?=[A-Z])");
  name = name.split(beforeCapitalLetter).map((e) => e.toLowerCase()).join(' ');
  return "No ${name}s found!";
}

String _errorTextGetter(dynamic error) => 'Something went wrong!';

/// The config for [PaginatedItemsBuilder].
class PaginatedItemsBuilderConfig {
  PaginatedItemsBuilderConfig({
    ShimmerConfig? shimmerConfig,
    this.mockItemGetter = _getByType,
    this.noItemsTextGetter = _noItemsTextGetter,
    this.errorTextGetter = _errorTextGetter,
    this.noItemsTextStyle = const TextStyle(
      fontWeight: FontWeight.w600,
      fontSize: 14,
    ),
    this.logErrors = true,
  }) : shimmerConfig = shimmerConfig ?? ShimmerConfig.defaultShimmer();

  /// Default config
  PaginatedItemsBuilderConfig.defaultConfig() {
    shimmerConfig = ShimmerConfig.defaultShimmer();
    mockItemGetter = _getByType;
    noItemsTextGetter = _noItemsTextGetter;
    errorTextGetter = _errorTextGetter;
    noItemsTextStyle = const TextStyle(
      fontWeight: FontWeight.w600,
      fontSize: 14,
    );
    logErrors = true;
  }

  /// Create a function and pass the reference to this.
  ///
  /// The function will be passed the type `T` and expects an object with mock
  /// data of type `T`. This calls the [PaginatedItemsBuilder]'s [itemBuilder]
  /// with the mockItem and overlays a shimmer for loading animation.
  ///
  /// You can also return a widget from this method, then that widget
  /// will be built in place of the loader.
  /// The widget is wrapped in an [IgnorePointer] when built to disable any
  /// onTap gesture listeners.
  ///
  /// ```dart
  /// class MockItems {
  ///   static dynamic getByType<T>([String? mockItemKey]) {
  //     final key = mockItemKey ?? T.toString();
  //     switch (key) {
  //       case 'Post':
  //         return _post;
  //     }
  //   }
  ///
  ///   static final _post = Post();
  ///   ```
  /// }
  late final dynamic Function<T>([String? key]) mockItemGetter;

  /// Provide a [ShimmerConfig] to customize the shimmer loading animation
  /// color, or the duration.
  late final ShimmerConfig shimmerConfig;

  /// Customize the text that is rendered when there are no items to display.
  late final String Function(String name) noItemsTextGetter;

  /// Customize the text that is rendered when an occurs.
  late final String Function(dynamic error) errorTextGetter;

  /// Customize the style of the text that is rendered when there
  /// are no items to display.
  late final TextStyle noItemsTextStyle;

  /// Whether to log errors to the console or not.
  late final bool logErrors;
}

/// [ShimmerConfig] class to customize the loading shimmer colors, duration etc.
class ShimmerConfig {
  /// The shimmer's base color. Defaults to Colors.grey[300].
  late final Color baseColor;

  /// The shimmer's highlight color. Defaults to Colors.grey[200].
  late final Color highlightColor;

  /// The shimmer's duration. Defaults to 800ms.
  late final Duration duration;

  /// The shimmer's direction. Defaults to [ShimmerDirection.ltr].
  late final ShimmerDirection direction;

  static final _defaultBaseColor = Colors.grey[300]!;
  static final _defaultHighlightColor = Colors.grey[200]!;
  static const _defaultDirection = ShimmerDirection.ltr;
  static const _defaultDuration = Duration(milliseconds: 800);

  ShimmerConfig({
    Color? baseColor,
    Color? highlightColor,
    this.direction = _defaultDirection,
    this.duration = _defaultDuration,
  })  : baseColor = baseColor ?? _defaultBaseColor,
        highlightColor = highlightColor ?? _defaultHighlightColor;

  /// default
  ShimmerConfig.defaultShimmer() {
    baseColor = _defaultBaseColor;
    highlightColor = _defaultHighlightColor;
    direction = _defaultDirection;
    duration = _defaultDuration;
  }
}
